(* FNCGrid Helper Class ******
   Description: This helper class will set the properties of the FNCGrid to
   preset values.

   Display Note: use TABs and TAB = 4

   TGridType:
   gtNormal = Normal size grid
   gtBig = The Font and height of each row is larger for a touch screen.
   gtStacked = The height of each row is even larger to allow to data fields
			to be stacked (normal font).

   To use drop a TTMSFNCGrid on a form, most properties are set in code by this
   helper class.  You may want to connect some Click events.  Include this unit in
   the form that contains the grid.  Add code like below to the unit:

   var
	aSettings : TGridSettings;
   begin
		aSettings.Create(TGridType.gtBig);
		aSettings.DataSource := dsPaidOutType;
		gridData.PrepareGrid(aSettings);
		gridData.AddColumn('Name', 18, 'Name Title');    // optional if multiple columns are wanted
		gridData.AddColumn('Code', 9, 'Code Title');     // optional
		gridData.Active := true;

   end;

   A call to ResizeColumns should be placed in the host form's OnResize event.

   The desired properties can be set in the TGridSettings record.  Note: aSettings
   is a record, there is no need to Free it.

   SORTING
   If column sorting is true, then you must add an OnColumnSorted event to the FNCGrid.
   Example:

   procedure TformMain.gridDataColumnSorted(Sender: TObject; ACol: Integer; Direction: TTMSFNCGridSortDirection);
   var
		sSortField : string;
   begin
		sSortField := gridData.FieldNameByColumnIndex(aCol);
		RunQuery(sSortField, Direction = TTMSFNCGridSortDirection.sdAscending);
   end;

   RESIZING COLUMNS
   The form that contains the FNCGrid needs to have a call to grid.ResizeColumns in the OnResize
   event of the form.

   GET FIELD NAME

   GET FIELD OBJECT
   Use SyncDBegin and SyncDEnd as pairs, you must call SyncDEnd if you have used SyndDBegin
   Example:

	aField := gridData.SyncDBegin(aCol, aRow, nActiveRecord);
	if aField <> nil then
	begin
		try
			if aField.FieldName = 'Total Amount' then
			begin
				if qryData.fieldbyname('Payment Status').AsString = 'Paid' then
					ALayout.Fill.Color := gcGreen;
			end;
		finally
			gridData.SyncDEnd(nActiveRecord);
		end;
	end;

	********************
	MULTI-SELECT Example BEGIN

	Grid Settings:  	gridData.Appearance.SelectedLayout.Fill.Color := gridData.Color;

	Use the TinSelectedItems class with the SelectItem and IsSelecte methods.
	Add OnSelectedCell and GetCellLayout events for the grid following the example below.

	procedure TformMain.gridDataSelectedCell(Sender: TObject; ACol, ARow: Integer);
	begin
		if FClickedRecord then  // This is set true in the OnClick event;
		begin
			SelectItem(tblData.FieldByName('Number').AsInteger);
			FClickedRecord := false;
			Button2.SetFocus;   // set the focus to some other control on the form.
		end;
	end;


	procedure TformMain.gridDataGetCellLayout(Sender: TObject; ACol, ARow: Integer; ALayout: TTMSFNCGridCellLayout; ACellState: TTMSFNCGridCellState);
	var
		aField : TField;
		nActiveRecord, nColumn : Integer;
	begin
		aField := gridData.SyncDBegin(aCol, aRow, nActiveRecord);
		if aField <> nil then
		begin
			try
				if IsSelected(tblData.FieldByName('Number').AsInteger) > -1 then
				begin
					ALayout.Fill.Color := clMoneyGreen;
					ALayout.Font.Color := clBlack;
				end
				else
				begin
					ALayout.Fill.Color := gridData.Color;
					ALayout.Font.Color := clBlack;
				end;

			finally
				gridData.SyncDEnd(nActiveRecord);
			end;
		end;
	end;
	MULTI-SELECT END *********************


   Copyright  2020 Inborne Technology
   10/19/2020, David - created unit
****************************************************************************)
unit FNCGridHelper;

interface

uses
	Vcl.Graphics,  // for clHighlight
	Data.DB,   // for TDatasource
	System.SysUtils,  // for Exception
	System.Generics.Collections,  // for TList
	VCL.TMSFNCTypes, VCL.TMSFNCUtils, VCL.TMSFNCGraphics,
	VCL.TMSFNCGraphicsTypes, VCL.TMSFNCGridCell, VCL.TMSFNCGridOptions, VCL.TMSFNCCustomComponent,
	VCL.TMSFNCCustomGrid, VCL.TMSFNCGridDatabaseAdapter, VCL.TMSFNCCustomControl,
	VCL.TMSFNCCustomScrollControl, VCL.TMSFNCGridData, VCL.TMSFNCGrid;

type
  TGridType = (gtNormal, gtBig, gtStacked, gtStacked3);

  TTMSFNCGridDataBaseAdapterOpen = class(TTMSFNCGridDatabaseAdapter);

  TGridSettings = record
	FontSize : integer;         // set by constructor
	FontName : string;          // set by constructor
	UseAlternatingRowColors : boolean;     // set by constructor
	OneDataField : string; 		 	// use this only if one visible field is wanted
	DataSource : TDatasource;     	// required
	RowHeight : integer;           	// set by constructor
	TitleRowHeight : integer;      	// set by constructor
	ShowHorzBorders : boolean;     	// set by constructor
	ShowVertBorders : boolean;     	// set by constructor
	AutoCreateColumns : boolean;   	// auto create columns from the data source
	AllowColumnSorting : boolean;   // Allow sorting by clickin on the title
	TouchScrolling : boolean;    	// set by constructor
  public
	constructor Create(nType : TGridType);
  end;

{ TinSelectedItems ************************************}
  TinSelectedItems = class(TObject)
  private
	FItemList : TList<integer>;
  public
    constructor Create;
	destructor Destroy; override;

	procedure Clear;
	function SelectItem(nValue : integer; boDeSelect : boolean = true) : integer;
    procedure DeSelectItem(nValue : integer);
	function IsSelected(nValue : integer) : integer;
	function ItemCount : integer;

	property Items : TList<integer> read FItemList;
  end;


  TFNCGridHelper = class helper for TTMSFNCGrid
  private
    function GetActive: boolean;
    procedure SetActive(const Value: boolean);
	function ExtractFieldNameFromHTML(const sHTMLValue: string): string;
	function GetGridFontSize: integer;
    procedure SetGridFontSize(const Value: integer);
  public
	function AddColumn(const sFieldName : string; nWidthPercent : integer; sColumnTitle : string = '') : TTMSFNCGridDatabaseAdapterColumn;
	procedure ClearColumns;
	procedure PrepareGrid(aSettings : TGridSettings);   // must be called first before other methods
	procedure ResizeColumns;

	function FieldNameByColumnIndex(const nCol : integer) : string;
	function FieldByColumnAsString(aCol, aRow : integer; var sFieldName, sFieldValue : string) : boolean;
	function SyncDBegin(const nCol, nRow : integer; var nActiveRecord : integer) : TField;
	procedure SyncDEnd(nActiveRecord : integer);

	property Active : boolean read GetActive write SetActive;  // must be set True AFTER all properties are set
    property GridFontSize : integer read GetGridFontSize write SetGridFontSize;
  end;



implementation

{ TFNCGridHelper }

// nWidth and sColumnTitle are optional
// the Tag is used to store the request percent for the column width
// the HTML tag <br> (lowercase) is used when there is more than one field and the font needs to be smaller
function TFNCGridHelper.AddColumn(const sFieldName : string; nWidthPercent : integer; sColumnTitle : string) : TTMSFNCGridDatabaseAdapterColumn;
var
	aDBAdapter : TTMSFNCGridDatabaseAdapter;
	aGridColumn : TTMSFNCGridColumn;
begin
	if self.Adapter = nil then
		raise Exception.Create('PrepareGrid must be called first.');

	aDBAdapter := TTMSFNCGridDatabaseAdapter(self.Adapter);

	// create the column in the grid and the DB adapter
	aGridColumn := self.Columns.Add;
	result := aDBAdapter.Columns.Add;    // manually create the columns

	if Pos('<#', sFieldName) > 0 then  // use HTML template for stacked fields
	begin
		if Pos('<br>', sFieldName) = 0 then   // if only one field then make the font larger
			result.HTMLTemplate := '<FONT size="18">' + sFieldName + '</FONT>'
		else
			result.HTMLTemplate := sFieldName;

		result.FieldName := ExtractFieldNameFromHTML(sFieldName);
	end
	else
		result.FieldName := sFieldName;

	if nWidthPercent = 0 then
		self.Columns[result.index].Tag := result.Field.DisplayWidth
	else
		self.Columns[result.index].Tag := nWidthPercent;

	if sColumnTitle <> '' then
		result.Header := sColumnTitle;
end;


procedure TFNCGridHelper.ClearColumns;
begin
	TTMSFNCGridDatabaseAdapter(self.Adapter).Columns.Clear;
	self.Columns.Clear;
end;


function TFNCGridHelper.GetActive: boolean;
begin
	result := self.Adapter.Active;
end;

function TFNCGridHelper.GetGridFontSize: integer;
begin
	result := self.Appearance.NormalLayout.Font.Size;
end;

// This function will extract the first field name from TMS HTMLTemplate string
function TFNCGridHelper.ExtractFieldNameFromHTML(const sHTMLValue : string) : string;
var
	nPos : integer;
begin
	result := '';
	nPos := Pos('#', sHTMLValue);
	if nPos > 1 then
	begin
		result := Copy(sHTMLValue, nPos + 1, Length(sHTMLValue) - nPos);
		nPos := Pos('>', result);
		if nPos > 1 then
			result := Copy(result, 1, nPos - 1)
		else
			result := '';
	end;
end;


function TFNCGridHelper.FieldNameByColumnIndex(const nCol: integer): string;
begin
	if self.Adapter <> nil then
	begin
		result := TTMSFNCGridDatabaseAdapter(self.Adapter).Columns[nCol - self.FixedColumns].FieldName;
		if result = '' then
			result := ExtractFieldNameFromHTML(TTMSFNCGridDatabaseAdapter(self.Adapter).Columns[nCol - self.FixedColumns].HTMLTemplate);
	end;
end;


// return: true = has a field object and returns the field name and the field value as a string
function TFNCGridHelper.FieldByColumnAsString(aCol, aRow : integer; var sFieldName, sFieldValue: string): boolean;
var
	aField : TField;
	aDBAdapter : TTMSFNCGridDataBaseAdapterOpen;
	nActiveRecord, nColumn : Integer;
begin
	result := false;
	sFieldName := '';
	sFieldValue := '';

	if (aRow > self.FixedRows - 1) then
	begin
		nColumn := aCol - self.FixedColumns;

		if (nColumn >= 0) and (nColumn <= self.Columns.Count - 1) then
		begin
			aDBAdapter := TTMSFNCGridDataBaseAdapterOpen(self.Adapter);
			if Assigned(aDBAdapter) then
			begin
				if aDBAdapter.CheckDataSet then
				begin
					nActiveRecord := aDBAdapter.DataLink.ActiveRecord;
					try
						aDBAdapter.SetActiveRecord(aRow);
						aField := aDBAdapter.Columns[nColumn].Field;

						if Assigned(aField) then
						begin
							sFieldName := aField.FieldName;
							sFieldValue := aField.AsString;
							result := true;
						end;
					finally
						aDBAdapter.DataLink.ActiveRecord := nActiveRecord;
					end;
				end;
			end;
		end;
	end;
end;


procedure TFNCGridHelper.PrepareGrid(aSettings : TGridSettings);
var
	aDBAdapter : TTMSFNCGridDatabaseAdapter;
	aColumn : TTMSFNCGridDatabaseAdapterColumn;
begin
	if self.Adapter = nil then
	begin
		aDBAdapter := TTMSFNCGridDatabaseAdapter.Create(self.Owner);
		self.Adapter := aDBAdapter;
	end
	else
		aDBAdapter := TTMSFNCGridDatabaseAdapter(self.Adapter);

	aDBAdapter.Grid := self;

	// True = create all the columns from the dataset; false = you must AddColumns manually
	aDBAdapter.AutoCreateColumns := aSettings.AutoCreateColumns;
	if not aSettings.AutoCreateColumns then
		aDBAdapter.AutoRemoveColumns := false;

	aDBAdapter.Datasource := aSettings.DataSource;

	self.Options.Selection.Mode := smSingleRow;  // highlight/select rows (not cells)
	self.Options.Editing.Enabled := false;       // no editing in the grid

	self.Appearance.SelectedLayout.Fill.Color := clHighlight;
	self.Appearance.BandLayout.Fill.Color := gcHoneydew;   // alternate row color
	self.Appearance.FocusedLayout.Fill.Color := self.Appearance.SelectedLayout.Fill.Color;
	self.Appearance.ShowFocus := false;
	self.AllowFocus := true;  // disable the grid showing a "focused" cell

	self.UseColumns := false;  // This allows the Fonts under Appearance to work
	GridFontSize := aSettings.FontSize;

	self.Options.Bands.Enabled := aSettings.UseAlternatingRowColors;
	self.Options.Mouse.TouchScrolling := aSettings.TouchScrolling;

	if aSettings.ShowHorzBorders and aSettings.ShowVertBorders then
	begin
		self.Options.Borders.CellBorders := TTMSFNCGridBorders.bAll;
		self.Options.Borders.FixedCellBorders := TTMSFNCGridBorders.bAll;
	end
	else
	if aSettings.ShowVertBorders then
	begin
		self.Options.Borders.CellBorders := TTMSFNCGridBorders.bVertical;
		self.Options.Borders.FixedCellBorders := TTMSFNCGridBorders.bVertical;
	end
	else
	if aSettings.ShowHorzBorders then
	begin
		self.Options.Borders.CellBorders := TTMSFNCGridBorders.bHorizontal;
		self.Options.Borders.FixedCellBorders := TTMSFNCGridBorders.bHorizontal;
	end
	else
	begin
		self.Options.Borders.CellBorders := TTMSFNCGridBorders.bNone;
		self.Options.Borders.FixedCellBorders := TTMSFNCGridBorders.bNone;
	end;

	if aSettings.AllowColumnSorting then
	begin
		self.Options.Sorting.Mode := TTMSFNCGridSortingMode.gsmNormal;
		self.SortColumn := 0;  // it is assumed that the first column in the default sort column
	end;

	self.ColumnCount := 0;
	if aSettings.OneDataField <> '' then
	begin
		self.ColumnCount := 1;
		aColumn := aDBAdapter.Columns.Add;    // manually create the columns
		aColumn.FieldName := aSettings.OneDataField;   // Dataset field name
		self.FixedRows := 0;     // no field header row needed
	end
	else
		self.FixedRows := 1;     // the fixed header row does not scroll

	self.FixedColumns := 0; // no fixed columnes
	if aSettings.RowHeight > 0 then
		self.DefaultRowHeight := aSettings.RowHeight;
	self.DefaultColumnWidth := 100;
	if aSettings.TitleRowHeight > 0 then
		self.RowHeights[0] := aSettings.TitleRowHeight
	else
		self.RowHeights[0] := 24;
end;


procedure TFNCGridHelper.ResizeColumns;
var
	nTotal : integer;
	nGridWidth : single;
	nIndex : integer;
begin
	nTotal := 0;
	nGridWidth := self.Width; // GetTotalColumnWidth;

	// get the total weighted values
	for nIndex := 0 to self.ColumnCount - 1 do
		nTotal := nTotal + self.Columns[nIndex].Tag;

	for nIndex := 0 to self.ColumnCount - 1 do
		self.ColumnWidths[nIndex] := (self.Columns[nIndex].Tag / nTotal) * nGridWidth;

	// When we have a Maximized window the columns don't size propertly unless we do the "..Update" thing
	self.BeginUpdate;
	self.EndUpdate;
end;


// The database adapter must be set active AFTER these settings are done,
// otherwise it doesn't do the Columns correctly
procedure TFNCGridHelper.SetActive(const Value: boolean);
begin
	self.Options.ColumnSize.Stretch := true;     // stretch the last column to fit the grid size; call only after all columns are added
//	self.Options.ColumnSize.StretchAll := True;  // stretch all columns to fit (Stretch must also be true)
	self.Adapter.Active := Value;
end;


procedure TFNCGridHelper.SetGridFontSize(const Value: integer);
begin
	if Value > 0 then
	begin
		self.Appearance.NormalLayout.Font.Size := Value;
		self.Appearance.SelectedLayout.Font.Size := Value;
		self.Appearance.FocusedLayout.Font.Size := Value;
		self.Appearance.BandLayout.Font.Size := Value;
		self.Appearance.FixedLayout.Font.Size := Value - 2;
	end;
end;


function TFNCGridHelper.SyncDBegin(const nCol, nRow : integer; var nActiveRecord : integer): TField;
var
	aDBAdapter : TTMSFNCGridDataBaseAdapterOpen;
	nColumn : Integer;
	sFieldName : string;
begin
	result := nil;
	nActiveRecord := -1;  // not set

	if self.Adapter <> nil then
	begin
		aDBAdapter := TTMSFNCGridDataBaseAdapterOpen(self.Adapter);
		if aDBAdapter.CheckDataSet then
		begin
			if aDBAdapter.Active and (nRow >= self.FixedRows) then
			begin
				nColumn := nCol - self.FixedColumns;
				if (nColumn >= 0) and (nColumn <= aDBAdapter.Columns.Count - 1) then
				begin
					nActiveRecord := aDBAdapter.DataLink.ActiveRecord;
					aDBAdapter.SetActiveRecord(nRow);

					// option 1
//					result := aDBAdapter.Columns[nColumn].Field;         // AV error after scrolling down

					// option 2
					sFieldName := aDBAdapter.Columns[nColumn].FieldName;             // AV error after scrolling down
					result := aDBAdapter.DataLink.DataSet.FieldByName(sFieldName);

					// option 3
//					result := aDBAdapter.DataLink.DataSet.Fields[nColumn]; // doesn't work, the display columns do not match the DataLink columns.
				end;
			end;
		end;
	end;
end;



procedure TFNCGridHelper.SyncDEnd(nActiveRecord : integer);
var
	aDBAdapter : TTMSFNCGridDataBaseAdapterOpen;
begin
	if nActiveRecord > -1 then
	begin
		aDBAdapter := TTMSFNCGridDataBaseAdapterOpen(self.Adapter);
		aDBAdapter.DataLink.ActiveRecord := nActiveRecord;
	end;
end;

{ TGridSettings }

constructor TGridSettings.Create(nType : TGridType);
begin
	if nType = TGridType.gtNormal then
	begin
		FontSize := 0;
		FontName := '';
		UseAlternatingRowColors := false;
		DataSource := nil;
		OneDataField := '';
		RowHeight := 0;
		TitleRowHeight := 0;
		ShowHorzBorders := true;
		ShowVertBorders := true;
		AutoCreateColumns := true;
		AllowColumnSorting := false;
		TouchScrolling := false;
	end
	else
	if nType = TGridType.gtBig then
	begin
		FontSize := 14;
		FontName := '';
		UseAlternatingRowColors := true;
		DataSource := nil;
		OneDataField := '';
		RowHeight := 34;
		TitleRowHeight := 24;
		ShowHorzBorders := false;
		ShowVertBorders := true;
		AutoCreateColumns := true;
		AllowColumnSorting := false;
		TouchScrolling := true;
	end
	else
	if nType = TGridType.gtStacked then
	begin
		FontSize := 10;
		FontName := '';
		UseAlternatingRowColors := true;
		DataSource := nil;
		OneDataField := '';
		RowHeight := 40;
		TitleRowHeight := 24;
		ShowHorzBorders := true;
		ShowVertBorders := true;
		AutoCreateColumns := false;
		AllowColumnSorting := false;
		TouchScrolling := true;
	end
	else
	if nType = TGridType.gtStacked3 then
	begin
		FontSize := 10;
		FontName := '';
		UseAlternatingRowColors := true;
		DataSource := nil;
		OneDataField := '';
		RowHeight := 57;
		TitleRowHeight := 24;
		ShowHorzBorders := true;
		ShowVertBorders := true;
		AutoCreateColumns := false;
		AllowColumnSorting := true;
		TouchScrolling := true;
	end;
end;


{ TinSelectedItems ***************************}

procedure TinSelectedItems.Clear;
begin
	FItemList.Clear;
end;

constructor TinSelectedItems.Create;
begin
	FItemList := TList<integer>.Create;
end;

destructor TinSelectedItems.Destroy;
begin
	if FItemList <> nil then
	begin
		FItemList.Clear;
		FItemList.Free;
	end;
	inherited;
end;

// returns the index number or -1 if not found
function TinSelectedItems.IsSelected(nValue : integer): integer;
var
	nIndex : integer;
begin
	result := FItemList.IndexOf(nValue);
end;

function TinSelectedItems.ItemCount: integer;
begin
	result := FItemList.Count;
end;

procedure TinSelectedItems.DeSelectItem(nValue: integer);
var
	nIndex, nLen : integer;
begin
	nIndex := IsSelected(nValue);
	if nIndex > -1 then
		FItemList.Delete(nIndex);
end;


function TinSelectedItems.SelectItem(nValue : integer; boDeSelect : boolean): integer;
var
	nCount : integer;
	nLen : integer;
begin
	result := IsSelected(nValue);
	if result = -1 then
		FItemList.Add(nValue);
end;


end.
